@@ -17,7 +17,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
  *
- * $Id: mms.c,v 1.16 2004/12/22 22:02:33 mathrick Exp $
+ * $Id: mms.c,v 1.20 2005/12/12 09:24:19 theuraeus Exp $
  *
  * MMS over TCP protocol
  *   based on work from major mms
@@ -61,8 +61,6 @@
 # define lprintf(x...)
 #endif
 
-#define LOG 
-
 
 #include "bswap.h"
 #include "mms.h"
@@ -144,8 +142,10 @@ struct mms_s {
   int           num_stream_ids;
   int           stream_ids[ASF_MAX_NUM_STREAMS];
   int           stream_types[ASF_MAX_NUM_STREAMS];
+  off_t         start_packet_seq; /* for live streams != 0, need to keep it around */
+  int           need_discont; /* whether we need to set start_packet_seq */
   int           asf_packet_len;
-  uint32_t      file_len;
+  uint64_t      file_len;
   char          guid[37];
   uint32_t      bitrates[ASF_MAX_NUM_STREAMS];
   uint32_t      bitrates_pos[ASF_MAX_NUM_STREAMS];
@@ -176,6 +176,8 @@ static off_t fallback_io_read(void *data, int socket, char *buf, off_t num)
   while (len < num)
   {
     ret = (off_t)read(socket, buf + len, num - len);
+    if(ret == 0)
+      break; /* EOF */
     if(ret < 0)
       switch(errno)
       {
@@ -381,7 +383,7 @@ static int get_guid (unsigned char *buffer, int offset) {
 
 static void print_command (char *data, int len) {
 
-#ifdef LOG
+#ifdef DEBUG
   int i;
   int dir = LE_32 (data + 36) >> 16;
   int comm = LE_32 (data + 36) & 0xFFFF;
@@ -436,22 +438,25 @@ static int send_command (mms_io_t *io, mms_t *this, int command,
   mms_buffer_init(&command_buffer, this->scmd);
   mms_buffer_put_32 (&command_buffer, 0x00000001);   /* start sequence */
   mms_buffer_put_32 (&command_buffer, 0xB00BFACE);   /* #-)) */
-  mms_buffer_put_32 (&command_buffer, length + 32);
+  mms_buffer_put_32 (&command_buffer, len8 * 8 + 32);
   mms_buffer_put_32 (&command_buffer, 0x20534d4d);   /* protocol type "MMS " */
   mms_buffer_put_32 (&command_buffer, len8 + 4);
   mms_buffer_put_32 (&command_buffer, this->seq_num);
   this->seq_num++;
   mms_buffer_put_32 (&command_buffer, 0x0);          /* timestamp */
   mms_buffer_put_32 (&command_buffer, 0x0);
-  mms_buffer_put_32 (&command_buffer, len8+2);
+  mms_buffer_put_32 (&command_buffer, len8 + 2);
   mms_buffer_put_32 (&command_buffer, 0x00030000 | command); /* dir | command */
   /* end of the 40 byte command header */
   
   mms_buffer_put_32 (&command_buffer, prefix1);
   mms_buffer_put_32 (&command_buffer, prefix2);
 
-  n = io_write(io,  this->s, this->scmd, length + CMD_HEADER_LEN + CMD_PREFIX_LEN);
-  if (n != (length + CMD_HEADER_LEN + CMD_PREFIX_LEN)) {
+  if (length & 7)
+	  memset(this->scmd + length + CMD_HEADER_LEN + CMD_PREFIX_LEN, 0, 8 - (length & 7));
+
+  n = io_write(io,  this->s, this->scmd, len8 * 8 + CMD_HEADER_LEN + CMD_PREFIX_LEN);
+  if (n != (len8 * 8 + CMD_HEADER_LEN + CMD_PREFIX_LEN)) {
     return 0;
   }
 
@@ -554,7 +559,7 @@ static int get_packet_header (mms_io_t *io, mms_t *this, mms_packet_header_t *he
   
 error:
   lprintf("read error, len=%d\n", len);
-  perror(NULL);
+  perror("Could not read packet header");
   return MMS_PACKET_ERR;
 }
 
@@ -715,15 +720,17 @@ static void interp_asf_header (mms_t *this) {
       case GUID_ASF_FILE_PROPERTIES:
 
         this->asf_packet_len = LE_32(this->asf_header + i + 92 - 24);
-        this->file_len   = LE_32(this->asf_header + i + 40 - 24);
+        this->file_len       = LE_64(this->asf_header + i + 40 - 24);
         lprintf ("file object, packet length = %d (%d)\n",
                  this->asf_packet_len, LE_32(this->asf_header + i + 96 - 24));
         break;
 
       case GUID_ASF_STREAM_PROPERTIES:
         {
+	  uint16_t flags;
           uint16_t stream_id;
-          int      type;  
+          int      type;
+          int      encrypted;
 
           guid = get_guid(this->asf_header, i);
           switch (guid) {
@@ -733,6 +740,8 @@ static void interp_asf_header (mms_t *this) {
               break;
     
             case GUID_ASF_VIDEO_MEDIA:
+	    case GUID_ASF_JFIF_MEDIA:
+	    case GUID_ASF_DEGRADABLE_JPEG_MEDIA:
               type = ASF_STREAM_TYPE_VIDEO;
               this->has_video = 1;
               break;
@@ -745,13 +754,20 @@ static void interp_asf_header (mms_t *this) {
               type = ASF_STREAM_TYPE_UNKNOWN;
           }
 
-          stream_id = LE_16(this->asf_header + i + 48);
+          flags = LE_16(this->asf_header + i + 48);
+	  stream_id = flags & 0x7F;
+	  encrypted = flags >> 15;
 
-          lprintf ("stream object, stream id: %d\n", stream_id);
-          this->stream_types[stream_id] = type;
-          this->stream_ids[this->num_stream_ids] = stream_id;
-          this->num_stream_ids++;
-      
+          lprintf ("stream object, stream id: %d, type: %d, encrypted: %d\n",
+		   stream_id, type, encrypted);
+
+	  if (this->num_stream_ids < ASF_MAX_NUM_STREAMS && stream_id < ASF_MAX_NUM_STREAMS) {
+	    this->stream_types[stream_id] = type;
+	    this->stream_ids[this->num_stream_ids] = stream_id;
+	    this->num_stream_ids++;
+	  } else {
+	    lprintf ("too many streams, skipping\n");
+	  }
         }
         break;
 
@@ -963,10 +979,12 @@ int static mms_choose_best_streams(mms_io_t *io, mms_t *this) {
       this->scmd_body [ (i - 1) * 6 + 7 ] = 0x00;
       
       /* forces the asf demuxer to not choose this stream */
-      this->asf_header[this->bitrates_pos[this->stream_ids[i]]]     = 0;
-      this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 1] = 0;
-      this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 2] = 0;
-      this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 3] = 0;
+      if (this->bitrates_pos[this->stream_ids[i]]) {
+        this->asf_header[this->bitrates_pos[this->stream_ids[i]]]     = 0;
+        this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 1] = 0;
+        this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 2] = 0;
+        this->asf_header[this->bitrates_pos[this->stream_ids[i]] + 3] = 0;
+      }
     }
   }
 
@@ -1018,6 +1036,8 @@ mms_t *mms_connect (mms_io_t *io, void *data, const char *url, int bandwidth) {
   this->asf_header_read = 0;
   this->num_stream_ids  = 0;
   this->asf_packet_len  = 0;
+  this->start_packet_seq= 0;
+  this->need_discont    = 1;
   this->buf_size        = 0;
   this->buf_read        = 0;
   this->has_audio       = 0;
@@ -1298,7 +1318,7 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
               lprintf ("new stream.\n");
               /* asf header */
               if (!get_asf_header (io, this)) {
-  /* FIXME: de-xine-ification */
+		/* FIXME: de-xine-ification */
                 lprintf ( "***LOG:*** -- "
 			 "failed to read new ASF header\n");
                 return 0;
@@ -1325,7 +1345,7 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
                 mms_buffer_put_8  (&command_buffer, 0x00);                        /* stream time limit flag */
                 mms_buffer_put_32 (&command_buffer, ASF_MEDIA_PACKET_ID_TYPE);    /* asf media packet id type */
                 if (!send_command (io, this, 0x07, 1, 0x0001FFFF, command_buffer.pos)) {
-  /* FIXME: de-xine-ification */
+		  /* FIXME: de-xine-ification */
                   lprintf ( "***LOG:*** -- "
 			   "libmms: failed to send command 0x07\n");
                   return 0;
@@ -1338,7 +1358,7 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
           case 0x1b:
             {
               if (!send_command (io, this, 0x1b, 0, 0, 0)) {
-  /* FIXME: de-xine-ification */
+		/* FIXME: de-xine-ification */
                 lprintf( "***LOG:*** -- "
 			"libmms: failed to send command\n");
                 return 0;
@@ -1350,7 +1370,7 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
             break;
   
           default:
-  /* FIXME: de-xine-ification */
+	    /* FIXME: de-xine-ification */
 	    lprintf ( "***LOG:*** -- "
                      "unexpected mms command %02x\n", command);
         }
@@ -1359,7 +1379,7 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
       break;
 
     case MMS_PACKET_ASF_HEADER:
-  /* FIXME: de-xine-ification */
+      /* FIXME: de-xine-ification */
       lprintf( "***LOG:*** -- "
 	      "libmms: unexpected asf header packet\n");
       this->buf_size = 0;
@@ -1369,21 +1389,30 @@ static int get_media_packet (mms_io_t *io, mms_t *this) {
       {
         /* media packet */
 
+	/* FIXME: probably needs some more sophisticated logic, but
+	   until we do seeking, this should work */
+	if(this->need_discont)
+	{
+	  this->need_discont = 0;
+	  this->start_packet_seq = header.packet_seq;
+	}
+	
         lprintf ("asf media packet detected, packet_len=%d, packet_seq=%d\n",
                  header.packet_len, header.packet_seq);
         if (header.packet_len > this->asf_packet_len) {
-  /* FIXME: de-xine-ification */
+	  /* FIXME: de-xine-ification */
 	  lprintf ( "***LOG:*** -- "
                    "libmms: invalid asf packet len: %d bytes\n", header.packet_len);
           return 0;
         }
     
         /* simulate a seek */
-        this->current_pos = (off_t)this->asf_header_len + (off_t)header.packet_seq * (off_t)this->asf_packet_len;
-    
+        this->current_pos = (off_t)this->asf_header_len +
+	  ((off_t)header.packet_seq - this->start_packet_seq) * (off_t)this->asf_packet_len;
+
         len = io_read(io,  this->s, this->buf, header.packet_len);
         if (len != header.packet_len) {
-  /* FIXME: de-xine-ification */
+	  /* FIXME: de-xine-ification */
           lprintf ( "***LOG:*** -- "
 		   "libmms: read failed\n");
           return 0;
